%% Lezione 9 Lab: K-means
% Scopo dell'esercitazione e' implementare e validare un algoritmo di
% clustering partizionale, il k-means.

clear all
close all
clc

%% Prima parte: costruzione del dataset

% Genero casualmente da tre gaussiane bidimensionali
mu1 = [1 1];
mu2 = [2 3];
mu3 = [-2 5];
sigma = [0.5 0; 0 0.5];

X1 = mvnrnd(mu1, sigma, 100);
X2 = mvnrnd(mu2, sigma, 100);
X3 = mvnrnd(mu3, sigma, 100);

data = [X1' X2' X3'];
scatter(data(1,:), data(2,:));

[nfeat, nobj] = size(data);

%% Implementazione: linee guida
clear C d

% Decido a priori il numero di cluster
n_cluster = 3;

% Inizializzazione
% (suggerimento: si parte da una clusterizzazione iniziale.
% Assegno ad ogni punto un etichetta a caso tra 1 e n_cluster.
% Controllare l'help del comando randi)
labels = randi(n_cluster,[1 nobj]);

% Continuo a ciclare; in ogni passo del ciclo:
% - Per ogni cluster, calcolo la media dei suoi punti (sto ri-stimando i
% centri di cluster)
% - Per ogni punto, calcolo la sua distanza (euclidea) da ogni centro di cluster, quindi
% ri-calcolo le etichette: ogni punto è assegnato al cluster più vicino.
% - Controllo se l'algoritmo è arrivato a convergenza:
% come: se le nuove etichette sono uguali a quelle del passo precedente.

iter = 0;
while true
         
    iter = iter+1;
    
    %Calcolo le medie dei cluster
    for k=1:n_cluster
        C(:,k) = mean(data(:,labels==k),2);
    end
    
    %Per ogni punto, calcolo la distanza da ogni cluster
    for k = 1:n_cluster
        d(k,:) = sqrt(sum((data - repmat(C(:,k), 1, nobj)).^2,1));
    end
    
    %Qualche plot....
    scatter(data(1,:), data(2,:), [], labels); hold on
    scatter(C(1,:), C(2,:), 200, '*r'); hold off
    title(['K-means clustering, K=' num2str(n_cluster) ', iter.' num2str(iter)]);
    pause(0.5)
    drawnow
    
    %Per ogni punto, trovo il cluster piu' vicino
    old_labels = labels;
    [~,labels] = min(d,[],1);
          
    %Check per la convergenza
    if labels == old_labels
        break
    end   
end

%%
% Ripetere l'esercizio con un dataset monodimensionale, in particolar modo
% l'immagine "T1.png" precaricata in matlab (importabile con il comando 
% I = im2double(imread('T1.png'));
% Le feature sono i livelli di grigio, gli oggetti da classificare sono i
% singoli pixel. Con 3 cluster, si riescono a separare le regioni
% del cervello?

clear C d
I = im2double(imread('T1.png'));
figure
subplot(1,2,1), imshow(I);
data = I(:)';
[nfeat, nobj] = size(data);

n_cluster = 3;
labels = randi(n_cluster,[1 nobj]);

subplot(1,2,2);
iter = 0;
while true
         
    iter = iter+1;
    
    %Calcolo le medie dei cluster
    for k=1:n_cluster
        C(:,k) = mean(data(:,labels==k),2);
    end
    
    %Per ogni punto, calcolo la distanza da ogni cluster
    for k = 1:n_cluster
        d(k,:) = sqrt(sum((data - repmat(C(:,k), 1, nobj)).^2,1));
    end
    
    %Qualche plot....
    imshow(reshape(labels,size(I)),[]); colormap jet
    title(['K-means clustering, K=' num2str(n_cluster) ', iter.' num2str(iter)]);
    pause(0.5)
    drawnow
    
    %Per ogni punto, trovo il cluster piu' vicino
    old_labels = labels;
    [~,labels] = min(d,[],1);
          
    %Check per la convergenza
    if labels == old_labels
        break
    end   
end
